/* DESCRIPTION:
 *	This is an RPC shuffler for passing NFS RCP's via TCP.
 *
 * AUTHOR:
 *	Simon J. Gerraty <sjg@zen.void.oz.au>
 */
/*
 *      @(#)Copyright (c) 1994, Simon J. Gerraty.
 *      
 *      Permission to use, modify and distribute this source code 
 *      is granted subject to the following conditions.
 *      1/ that the above copyright notice and this notice 
 *      are preserved in all copies and that due credit be given 
 *      to the author.  
 *      2/ that any changes to this code are clearly commented 
 *      as such so that the author does not get blamed for bugs 
 *      other than his own.
 *      
 *      Please send copies of changes and bug-fixes to:
 *      sjg@zen.void.oz.au
 */

#ifndef lint
static char RCSid[] = "$Id: unfsclnt.c,v 1.10 1995/05/13 07:29:43 sjg Exp $";
#endif
#include <stdio.h>
#include <sys/param.h>
#include <sys/mount.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/stat.h>
#include <sys/syslog.h>

#include <rpc/rpc.h>
#include <sys/types.h>
#include <errno.h>

#include <err.h>
#include <fstab.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "unfsd.h"
#include "mount_prot.h"
#include "nfs_prot.h"

#if defined(BSD4_4)
# define HAVE_SETSID
#endif

/* not all RPC headers declare this */
SVCXPRT *svcfd_create	__P((int fd, unsigned int sendsize, unsigned int recvsize));

static CLIENT *MntClp;
static CLIENT *NfsClp;
static struct sockaddr_in Dest;

static int WantExit = 0;
static int UnmountDone = 0;
static struct timeval TIMEOUT = { 25, 0 };


static int
do_unmount()
{
	int sts = 0;

	if (UnmountDone <= 0) {
		syslog(LOG_DEBUG, "unmount(%s)...", UnfsMntPt);
		if ((sts = unmount(UnfsMntPt, 0)) == 0) {
			uclnts_delete(UnfsMntSpec, UnfsMntPt);
			UnmountDone = 1;
		}
	}
	return sts;
}

static void
cleanup()
{
	if (WantExit++ == 0 && UnmountDone == 0) {
		char res = 0;
		
		/* do an umnt RPC so server will close down */
		syslog(LOG_DEBUG, "sending  MOUNTPROC_UMNT %s...",
		       UnfsMntedDir);
		clnt_call(MntClp, MOUNTPROC_UMNT,
			  xdr_dirpath, &UnfsMntedDir,
			  xdr_void, &res, TIMEOUT);
		do_unmount();
	}
	syslog(LOG_DEBUG, "exiting...");
}

static void
sigcatcher(sig)
	int sig;
{
	switch (sig) {
	case SIGALRM:
		if (WantExit == 0) {
			char res = 0;
			
			/* ping server */
			if (clnt_call(NfsClp, NFSPROC_NULL,
				      xdr_void, &res,
				      xdr_void, &res,
				      TIMEOUT) == RPC_SUCCESS) {
				alarm(_rpcsvc_closedown);
				return;
			}
		} else if (_rpcsvcdirty == 0) {
#ifndef NO_FWTK
			if ((UnfsFlags & UNFS_AUTH) && do_unmount() != 0) {
				alarm(5);		/* try again soon */
				return;
			}
#endif
			exit(0);
		}
		(void) alarm(_rpcsvc_closedown);
		return;
		break;
	case SIGTERM:
		exit(0);
	default:
		exit(1);
	}
}

#if DO_UNFS_PROG
#ifndef NO_FWTK
/*
 *	If UNFS_AUTH is set, unfsd cannot simply go away, or we will
 *	continually have to re-mount.
 *	So, he polls us to check that we are still here.
 *	He also calls us to say goodbye...
 *
 *	REVISIT: currently causes deadlock.  Using separate socket may
 *	well solve it - needs to be tested...
 */

void *
unfs_null(argp, rqstp)
void *argp;
struct svc_req *rqstp;
{
	static char res;

	res = 0;
#ifdef DEBUG
	syslog(LOG_DEBUG, "ping...");
#endif
	return &res;
}

void
unfs_prog_1(rqstp, transp)
	struct svc_req *rqstp;
	SVCXPRT *transp;
{
	void *(*local)();
	union {
		char char_arg;
	} argument;
	char *result = NULL;
	bool_t (*xdr_argument)(), (*xdr_result)();
	int old_dirty = _rpcsvcdirty;
	
	_rpcsvcdirty = 1;
	switch (rqstp->rq_proc) {
	case UNFS_QUIT:
		syslog(LOG_INFO, "server requested shutdown");
		WantExit++;			/* skip unmount RPC */
		_rpcsvcdirty = old_dirty = 0;	/* he may be gone by now */
		alarm(5);			/* FALLTHROUGH */
	case UNFS_NULL:
		xdr_argument = xdr_void;
		xdr_result = xdr_void;
		local = unfs_null;
		break;
	default:
		svcerr_noproc(transp);
		_rpcsvcdirty = 0;
		return;
	}
	bzero((char *)&argument, sizeof(argument));
	if (!svc_getargs(transp, xdr_argument, &argument)) {
                svcerr_decode(transp);
                return;
        }
	result = (*local)(&argument, rqstp);
        if (result != NULL && !svc_sendreply(transp, xdr_result, result)) {
                svcerr_systemerr(transp);
        }
        if (!svc_freeargs(transp, xdr_argument, &argument)) {
                _msgout(stderr, "unable to free arguments");
                exit(1);
        }
	_rpcsvcdirty = old_dirty;
        return;
}
#endif
#endif

/*
 *	Called when we lose connection to server
 */
int
unfsc_reconnect()
{
	int sock = RPC_ANYSOCK;
	char *server = ip_caller_id(&Dest, 0);

#ifndef NO_FWTK
	if ((UnfsFlags & UNFS_AUTH) && WantExit++ == 0) {	/* can't do it */
		/*
		 * we need to try and get kernel to give up on this
		 * RPC, so we can unmount filesystem...
		 */
		alarm(5);
	}
#endif
	if (WantExit != 0)
		return 0;
	
	if (MntClp) {
		clnt_destroy(MntClp);
		MntClp = NULL;
	}
	if (NfsClp) {
		clnt_destroy(NfsClp);
		NfsClp = NULL;
	}
	syslog(LOG_INFO, "reconnecting to %s", server);
	while (MntClp == NULL) {
		MntClp = clnttcp_create(&Dest, MOUNTPROC_MNT,
					MOUNTVERS,
					&sock, 0, 0);
		if (MntClp == NULL) {
			syslog(LOG_WARNING, "%s, still trying...",
			       clnt_spcreateerror(server));
			sleep(60);
		}
	}
	NfsClp = clnttcp_create(&Dest, 
				NFS_PROGRAM, NFS_VERSION,
				&sock, 0, 0);
	MntClp->cl_auth = authunix_create_default();
	NfsClp->cl_auth = authunix_create_default();
	return 1;
}

/*
 * This is just to centralize error handling.
 */

void
do_rpc(rqstp, transp, cl, xdr_argument, argument, xdr_result, result)
	struct svc_req *rqstp;
	SVCXPRT *transp;
	CLIENT *cl;
	xdrproc_t xdr_argument, xdr_result;
	void *argument, *result;
{
	int res;

	/*
	 * unpack the args...
	 * to avoid doing this, we'd need to write our own xdr
	 * routines that simply memcpy'd the arg as is into the
	 * outgoing message.
	 * Or we could build our own clnt*_call routines.
	 * Bottom line is that this works and will do for now.
	 */
	
	if (!svc_getargs(transp, xdr_argument, argument)) {
		svcerr_decode(transp);
		_rpcsvcdirty = 0;
		return;
	}

	/*
	 * now pack it up again and send to server...
	 */
	if ((res = clnt_call(cl, rqstp->rq_proc,
		      xdr_argument, argument,
		      xdr_result, result, TIMEOUT)) == RPC_SUCCESS) {
		if (!svc_sendreply(transp, xdr_result, result))
			svcerr_systemerr(transp);
	} else {
		syslog(LOG_ERR, "RPC failure (%d): %s", res,
		       clnt_sperror(cl, UnfsMntSpec));
		
		/*
		 *	See rpc/clnt.h for defns of error status values. This
		 *	list may need some additions.
		 */
		switch (res)
		{
		case RPC_CANTSEND:
		case RPC_CANTRECV:
		case RPC_TIMEDOUT:
			/*
			 *	Server died?  Try to re-establish.
			 */
			unfsc_reconnect();
			break;
		case RPC_CANTENCODEARGS:
			/*
			 *	Retrying is not going to help.
			 *	The data is bogus.
			 */
			break;
			/*
			 *	These should not happen, and if they do,
			 *	retrying won't help. Wait for
			 *	programmer to fix. 
			 */
		case RPC_VERSMISMATCH:
		case RPC_UNKNOWNHOST:
#ifdef RPC_UNKNOWNADDR
		case RPC_UNKNOWNADDR:
#endif
			exit(1);
			break;

		case RPC_UNKNOWNPROTO:
			/*
			 *	This is even worse :-)
			 */
			abort();
			break;
		}
	}

	/*
	 * now just clean up...
	 */
	if (!svc_freeargs(transp, xdr_argument, argument)) {
		_msgout("unable to free arguments");
		exit(1);
	}
	if (!svc_freeargs(transp, xdr_result, result)) {
                _msgout("unable to free result");
                exit(1);
        }

	_rpcsvcdirty = 0;
	return;
}

void
nfs_bounce(rqstp, transp)
	struct svc_req *rqstp;
	SVCXPRT *transp;
{
	/*
	 * Since we don't actually use any of the args, these
	 * unions, simply reserve enough space to unpack them.
	 */
	union {
		nfs_fh 		fh_arg;
		sattrargs 	sa_arg;
		diropargs 	dop_arg;
		readargs 	read_arg;
		writeargs 	write_arg;
		createargs 	create_arg;
		renameargs 	rename_arg;
		linkargs 	link_arg;
		symlinkargs 	slink_arg;
		readdirargs 	readdir_arg;
	} argument;
	union {
		char 		ch_res;
		attrstat 	at_res;
		diropres 	dir_res;
		nfsstat 	stat_res;
		readdirres 	rdir_res;
		readlinkres 	rlnk_res;
		readres 	read_res;
		statfsres 	sfs_res;
	} result;
	
	bool_t (*xdr_argument)(), (*xdr_result)();

	alarm(_rpcsvc_closedown);		/* reset idle timeout */
	_rpcsvcdirty = 1;
#ifdef DEBUG
	if (UnfsFlags & UNFS_DEBUG)
		syslog(LOG_DEBUG, "nfsprog_2(%d)",
		       rqstp->rq_proc);
#endif
	switch (rqstp->rq_proc) {
	case NFSPROC_NULL:
		xdr_argument = xdr_void;
		xdr_result = xdr_void;
		break;

	case NFSPROC_GETATTR:
		xdr_argument = xdr_nfs_fh;
		xdr_result = xdr_attrstat;
		break;

	case NFSPROC_SETATTR:
		xdr_argument = xdr_sattrargs;
		xdr_result = xdr_attrstat;
		break;

	case NFSPROC_ROOT:
		xdr_argument = xdr_void;
		xdr_result = xdr_void;
		break;

	case NFSPROC_LOOKUP:
		xdr_argument = xdr_diropargs;
		xdr_result = xdr_diropres;
		break;

	case NFSPROC_READLINK:
		xdr_argument = xdr_nfs_fh;
		xdr_result = xdr_readlinkres;
		break;

	case NFSPROC_READ:
		xdr_argument = xdr_readargs;
		xdr_result = xdr_readres;
		break;

	case NFSPROC_WRITECACHE:
		xdr_argument = xdr_void;
		xdr_result = xdr_void;
		break;

	case NFSPROC_WRITE:
		xdr_argument = xdr_writeargs;
		xdr_result = xdr_attrstat;
		break;

	case NFSPROC_CREATE:
		xdr_argument = xdr_createargs;
		xdr_result = xdr_diropres;
		break;

	case NFSPROC_REMOVE:
		xdr_argument = xdr_diropargs;
		xdr_result = xdr_nfsstat;
		break;

	case NFSPROC_RENAME:
		xdr_argument = xdr_renameargs;
		xdr_result = xdr_nfsstat;
		break;

	case NFSPROC_LINK:
		xdr_argument = xdr_linkargs;
		xdr_result = xdr_nfsstat;
		break;

	case NFSPROC_SYMLINK:
		xdr_argument = xdr_symlinkargs;
		xdr_result = xdr_nfsstat;
		break;

	case NFSPROC_MKDIR:
		xdr_argument = xdr_createargs;
		xdr_result = xdr_diropres;
		break;

	case NFSPROC_RMDIR:
		xdr_argument = xdr_diropargs;
		xdr_result = xdr_nfsstat;
		break;

	case NFSPROC_READDIR:
		xdr_argument = xdr_readdirargs;
		xdr_result = xdr_readdirres;
		break;

	case NFSPROC_STATFS:
		xdr_argument = xdr_nfs_fh;
		xdr_result = xdr_statfsres;
		break;
		
	default:
		svcerr_noproc(transp);
		_rpcsvcdirty = 0;
		return;
	}
	bzero((char *)&argument, sizeof(argument));
	bzero((char *)&result, sizeof(result));

	do_rpc(rqstp, transp, NfsClp,
	       xdr_argument, &argument,
	       xdr_result, &result);

	return;
}

void
mount_bounce(rqstp, transp)
	struct svc_req *rqstp;
	SVCXPRT *transp;
{
	union {
		dirpath 	dir_arg;
	} argument;
	union {
		char 		ch_res;
		fhstatus	fh_res;
	} result;
	bool_t (*xdr_argument)(), (*xdr_result)();

	alarm(_rpcsvc_closedown);		/* reset idle timeout */
	_rpcsvcdirty = 1;
#ifdef DEBUG
	if (UnfsFlags & UNFS_DEBUG)
		syslog(LOG_DEBUG, "mountprog_1(%d)",
		       rqstp->rq_proc);
#endif
	switch (rqstp->rq_proc) {
	case MOUNTPROC_NULL:
		xdr_argument = xdr_void;
		xdr_result = xdr_void;
		break;

	case MOUNTPROC_MNT:
		xdr_argument = xdr_dirpath;
		xdr_result = xdr_fhstatus;
		break;

	case MOUNTPROC_UMNT:
		xdr_argument = xdr_dirpath;
		xdr_result = xdr_void;
		uclnts_delete(UnfsMntSpec, UnfsMntPt);	/* clean this up too. */
		WantExit = 1;
		UnmountDone = 1;
		alarm(10);
		break;
		
	default:
		svcerr_noproc(transp);
		_rpcsvcdirty = 0;
		return;
	}
	bzero((char *)&argument, sizeof(argument));
	bzero((char *)&result, sizeof(result));

	do_rpc(rqstp, transp, MntClp,
	       xdr_argument, &argument,
	       xdr_result, &result);

	return;
}


/*
 * register a routine to receive the RPC's
 * it will simply grab them out of the UPD socket
 * and send them to the TCP one etc etc.
 */

int
unfsc(transp)
	SVCXPRT *transp;
{
	closelog();
	openlog("unfsc", LOG_PID, LOG_LOCAL1);
#if DO_UNFS_PROG
#ifndef NO_FWTK
	if (UnfsFlags & UNFS_AUTH)
		syslog(LOG_DEBUG, "registered UNFS control service.");
#endif
#endif
	atexit(cleanup);

	if (!svc_register(transp, MOUNTPROG, MOUNTVERS, mount_bounce, 0)) {
		_msgout("unable to register (MOUNTPROG, MOUNTVERS, udp).");
		exit(1);
	}
	if (!svc_register(transp, NFS_PROGRAM, NFS_VERSION, nfs_bounce, 0)) {
		_msgout("unable to register (NFS_PROGRAM, NFS_VERSION, tcp).");
		exit(1);
	}
	syslog(LOG_DEBUG, "registered UDP services");
	(void) signal(SIGALRM, sigcatcher);
	(void) signal(SIGTERM, sigcatcher);
	svc_run();
	_msgout("svc_run returned");
	exit(1);
	/* NOTREACHED */
}
	

/*
 * Open a UDP socket to listen on and
 * fork a child to run unfsc
 */
int
fork_unfsc(t, clp, sa, mntpt, dir)
	int t;
	CLIENT *clp;
	struct sockaddr_in *sa;
	char *mntpt, *dir;
{
	int u, port, len;
	struct sockaddr_in me;
	SVCXPRT *transp;

	_rpcsvc_closedown = _RPCSVC_CLOSEDOWN / 2; /* less than server */

#if DO_UNFS_PROG
#ifndef NO_FWTK
	SVCXPRT *unfs_transp;

	if (UnfsFlags & UNFS_AUTH) {
		unfs_transp = svcfd_create(t, 0, 0); /* already connected */
		/*
		 * we need this to authenticate caller
		 * see rendezvous_request(svc_tcp.c)
		 */
		unfs_transp->xp_raddr = *sa;
		unfs_transp->xp_addrlen = sizeof (struct sockaddr_in);
		if (!svc_register(unfs_transp,
				  UNFS_PROG, UNFS_VERS, unfs_prog_1, 0)) {
			_msgout("unable to register (UNFS_PROG, UNFS_VERS).");
			exit(1);
		}
	} else
		unfs_transp = NULL;
#endif
#endif
	MntClp = clp;
	
	/*
	 * need a new handle for NFS RPC's
	 */
	NfsClp = clnttcp_create(sa, 
				NFS_PROGRAM, NFS_VERSION,
				&t, 0, 0);
	NfsClp->cl_auth = authunix_create_default();
	/*
	 * save content of sa
	 */
	memcpy((char *) &Dest, (char *) sa, sizeof (struct sockaddr_in));
	
	/*
	 * now update sa with our address...
	 */
	if ((u = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
		syslog(LOG_ERR, "socket: %m");
		exit(1);
	}
	/*
	 * sadly there is no UDP equivalent of svcfd_create()
	 */
	transp = svcudp_create(u);
	
	/*
	 * need to update sa so kernel knows where we are...
	 */
	len = sizeof (struct sockaddr_in);
	if (getsockname(u, (struct sockaddr *)&me, &len) == 0) {
		memcpy((char *) sa, (char *) &me, sizeof (struct sockaddr_in));
		sa->sin_addr.s_addr = htonl(LOCALHOST_ADDR);	/* localhost */
	}
	
	port = ntohs(sa->sin_port);
	uclnts_delete(UnfsMntSpec, UnfsMntPt);

	if (UnfsFlags & UNFS_DEBUG)
		(void) fprintf(stderr, "unfsc: listening on: %s\n",
			       ip_caller_id(sa, 0));

	/*
	 * now fork so mount_unfs can exit
	 */
#ifndef HAVE_SETSID
	if (!(UnfsFlags & UNFS_ISBGRND))
		setpgrp(0, 0);
#endif
	switch (fork()) {
	case -1:
		syslog(LOG_ERR, "fork: %m");
		exit(1);
	case 0:
		if (!(UnfsFlags & UNFS_ISBGRND)) {
#ifdef HAVE_SETSID
			(void) setsid();
#endif
			(void) close(STDIN_FILENO);
			(void) close(STDOUT_FILENO);
			(void) close(STDERR_FILENO);
			(void) chdir("/");
		}
		break;
	default:
		/* parent wants port no... */
		return port;
	}
	uclnts_add(UnfsMntSpec, UnfsMntPt, port);
	unfsc(transp);
	exit(0);
}
